# load packages
import os
import sys
import json
import random as python_random
import numpy as np
import tensorflow as tf

import keras
from keras.layers import Input, Embedding, Dense, Bidirectional, Lambda, Multiply
from keras.layers import CuDNNLSTM, Conv1D, GlobalMaxPooling1D, GlobalAveragePooling1D
from keras.layers import Dropout, concatenate
from keras.models import Model, Sequential
from keras import backend as K

from utils import attend, classify, set_trainability, RandomAttention

argv = sys.argv
try:
    gpu_idx = eval(argv[1])
except IndexError:
    gpu_idx = 0

os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % gpu_idx
os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'

np.random.seed(0)
python_random.seed(0)
tf.set_random_seed(0)


def build_session(gpu_fraction=0.2, allow_growth=True):
    gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=gpu_fraction,
                                allow_growth=allow_growth)
    return tf.Session(config=tf.ConfigProto(gpu_options=gpu_options))


sess = build_session()
K.set_session(sess)
init_op = tf.group(tf.global_variables_initializer(),
                   tf.local_variables_initializer())
sess.run(init_op)

# load data
data = np.load('./data.npz')
X_kv = data['X_kv']
X_q = data['X_q']
y = data['y']
emb_matrix = data['emb']

kv_len_origin = data['kv_len_origin']
kv_len = X_kv.shape[1]
q_len = X_q.shape[1]
voab_size = emb_matrix.shape[0]
emb_dim = emb_matrix.shape[1]


# define the network

def get_model(params, emb_matrix):
    kv_len = params['kv_len']
    q_len = params['q_len']
    voab_size = params['voab_size']
    emb_dim = params['emb_dim']

    # encode data
    input_kv = Input(shape=(kv_len,), dtype='int32', name='Input_kv')
    input_q = Input(shape=(q_len,), dtype='int32', name='Input_q')

    emb = Embedding(voab_size, emb_dim, weights=[
                    emb_matrix], trainable=False, name='WordEmbedding')
    emb_kv = emb(input_kv)
    emb_q = emb(input_q)

    # attention
    assert params['use_attention'] == True, 'Parameter Error: use_attention'

    emb_kv_attention, mask, module_attention = attend(emb_q, emb_kv, params)
    preds_attention, module_classifier = classify(emb_kv_attention, params)

    emb_kv_pretrain = RandomAttention(
        keep_rate=1 / kv_len, attention_shape=(kv_len, 1))(emb_kv)
    preds_pretrain = module_classifier(emb_kv_pretrain)

    set_trainability(module_classifier, False)
    model_attention = Model(
        inputs=[input_kv, input_q], outputs=preds_attention)
    model_attention.compile(loss='binary_crossentropy',
                            optimizer='rmsprop',
                            metrics=['acc'])

    set_trainability(module_classifier, True)
    model_pretrain = Model(inputs=[input_kv, input_q], outputs=preds_pretrain)
    model_pretrain.compile(loss='binary_crossentropy',
                           optimizer='rmsprop',
                           metrics=['acc'])

    model_mask = Model(inputs=[input_kv, input_q], outputs=mask)
    model_mask.compile(loss='mse', optimizer='sgd')

    return model_attention, model_pretrain, model_mask, module_attention, module_classifier


# train the model
use_attention = True
for kv_encoder in ['pooling', 'rcnn']:
    params = {
        'use_attention': use_attention,
        'kv_encoder': kv_encoder,

        'kv_len': kv_len,
        'q_len': q_len,
        'voab_size': voab_size,
        'emb_dim': emb_dim,
    }
    print(params)

    model_attention, model_pretrain, model_mask, module_attention, module_classifier = get_model(
        params, emb_matrix)

    set_trainability(module_classifier, True)
    model_pretrain.summary()
    model_pretrain.fit([X_kv, X_q], y, batch_size=128, epochs=25)

    set_trainability(module_classifier, False)
    model_attention.summary()
    model_attention.fit([X_kv, X_q], y, batch_size=128, epochs=25)

    # evaluate the model
    result = model_attention.evaluate([X_kv, X_q], y, batch_size=128)
    print(result)
    masks = model_mask.predict([X_kv, X_q], batch_size=128)
    neg_masks = masks[y == 0]
    pos_masks = masks[y == 1]

    result_str = {
        'kv_encoder': str(kv_encoder),
        'kv_len': int(kv_len),
        'kv_len_origin': int(kv_len_origin),
        'train_loss': float(result[0]),
        'train_acc': float(result[1]),

        'pos_default': float(np.sum(pos_masks[:, kv_len_origin:]) / np.sum(pos_masks)),
        'pos_A': float(np.sum(pos_masks[:, kv_len_origin]) / np.sum(pos_masks)),
        'pos_B': float(np.sum(pos_masks[:, kv_len_origin + 1]) / np.sum(pos_masks)),
        'pos_C': float(np.sum(pos_masks[:, kv_len_origin + 2]) / np.sum(pos_masks)),
        'pos_D': float(np.sum(pos_masks[:, kv_len_origin + 3]) / np.sum(pos_masks)),

        'neg_default': float(np.sum(neg_masks[:, kv_len_origin:]) / np.sum(neg_masks)),
        'neg_A': float(np.sum(neg_masks[:, kv_len_origin]) / np.sum(neg_masks)),
        'neg_B': float(np.sum(neg_masks[:, kv_len_origin + 1]) / np.sum(neg_masks)),
        'neg_C': float(np.sum(neg_masks[:, kv_len_origin + 2]) / np.sum(neg_masks)),
        'neg_D': float(np.sum(neg_masks[:, kv_len_origin + 3]) / np.sum(neg_masks)),
    }
    print(result_str)

    f = open('./pretrain.jsonl', 'a')
    f.write(json.dumps(result_str) + '\n')
    f.close()
