# load packages
import os
import sys
import json
import numpy as np

import tensorflow as tf
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import random_ops

import keras
from keras.layers import Input, Embedding, Dense, Bidirectional, Lambda, Multiply
from keras.layers import CuDNNLSTM, LSTM, Conv1D, GlobalMaxPooling1D, GlobalAveragePooling1D
from keras.layers import Dropout, concatenate
from keras.models import Model, Sequential
from keras.utils import Sequence
from keras import backend as K


class DataGenerator(Sequence):

    def __init__(self, X_kv, X_q, y=None,
                 shuffle=False, batch_size=128):
        self.X_kv = X_kv
        self.X_q = X_q
        self.y = y

        self.n = X_kv.shape[0]

        self.shuffle = shuffle
        self.batch_size = batch_size

        self._set_index_array()

    def __getitem__(self, idx):
        batch_index = self.index_array[self.batch_size * idx:
                                       self.batch_size * (idx + 1)]
        batch_X = [self.X_kv[batch_index], self.X_q[batch_index]]

        if self.y is None:
            return (batch_X,)
        else:
            batch_y = self.y[batch_index]
            return (batch_X, batch_y)

    def __len__(self):
        return (self.n + self.batch_size - 1) // self.batch_size

    def on_epoch_end(self):
        self._set_index_array()

    def _set_index_array(self):
        if self.shuffle:
            self.index_array = np.random.permutation(self.n)
        else:
            self.index_array = np.arange(self.n)


def set_trainability(model, trainable=True):
    model.trainable = trainable
    for layer in model.layers:
        layer.trainable = trainable


class ConcatenateVecSeq(keras.layers.Layer):

    def __init__(self, **kwargs):
        super(ConcatenateVecSeq, self).__init__(**kwargs)

    def call(self, inputs):
        input1, input2 = inputs
        input1 = tf.tile(tf.expand_dims(input1, axis=-2),
                         [1, int(input2.get_shape()[1]), 1])
        return tf.concat([input1, input2], axis=-1)

    def compute_output_shape(self, input_shapes):
        input_shape1, input_shape2 = input_shapes
        input_shape = list(input_shape2)
        input_shape[-1] = int(input_shape[-1]) + int(input_shape1[-1])
        input_shape[-2] = int(input_shape[-2])
        return tuple(input_shape)


class RandomAttention(keras.layers.Layer):

    def __init__(self, keep_rate, attention_shape=None, seed=None, **kwargs):
        super(RandomAttention, self).__init__(**kwargs)
        self.keep_rate = min(1., max(0., keep_rate))
        self.attention_shape = attention_shape
        self.seed = seed

    def call(self, inputs):
        if self.attention_shape is None:
            attention_shape = K.shape(inputs)
        else:
            batch_size = K.shape(inputs)[0]
            attention_shape = tf.concat(
                [[batch_size], self.attention_shape], axis=0)

        random_tensor = random_ops.random_uniform(
            attention_shape, dtype=tf.float32) + self.keep_rate
        binary_tensor = math_ops.floor(random_tensor)
        outputs = inputs * binary_tensor

        return outputs

    def get_config(self):
        config = {'keep_rate': self.keep_rate,
                  'attention_shape': self.attention_shape,
                  'seed': self.seed}
        base_config = super(RandomAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape


def recurrent_convolutional_encoder(emb):
    hidden = Bidirectional(CuDNNLSTM(64, return_sequences=True))(emb)
    hidden = concatenate([hidden, emb])
    hidden = Conv1D(64, 3, padding='valid', activation='relu')(hidden)
    outputs = GlobalMaxPooling1D()(hidden)

    return outputs


def attend(emb_q, emb_kv, params):
    input_q_emb = Input(shape=(params['q_len'], params[
        'emb_dim']), dtype='float32')
    input_kv_emb = Input(shape=(params['kv_len'], params[
        'emb_dim']), dtype='float32')

    encode_q = recurrent_convolutional_encoder(input_q_emb)

    x = ConcatenateVecSeq()([encode_q, input_kv_emb])
    x = Dense(64, activation='relu')(x)
    x = Dense(1, activation=None)(x)
    x = Lambda(lambda x: K.softmax(x, axis=1))(x)

    encode_kv = Multiply()([x, input_kv_emb])
    mask = Lambda(lambda x: tf.squeeze(x, axis=-1))(x)

    attention_model = Model(
        inputs=[input_q_emb, input_kv_emb], outputs=[encode_kv, mask], name='Attention')
    attention_model.compile(loss='mse', optimizer='sgd')

    encode_kv, mask = attention_model([emb_q, emb_kv])

    return encode_kv, mask, attention_model


def classify(emb_kv, params):
    input_emb = Input(shape=(params['kv_len'], params[
        'emb_dim']), dtype='float32')

    if params['kv_encoder'] == 'rcnn':
        encode_kv = recurrent_convolutional_encoder(input_emb)
    elif params['kv_encoder'] == 'pooling':
        if params['use_attention']:
            encode_kv = Lambda(lambda x: tf.reduce_sum(x, axis=1))(input_emb)
        else:
            encode_kv = GlobalAveragePooling1D()(input_emb)
    else:
        raise NotImplementedError

    encode_kv = Dense(64, activation='relu')(encode_kv)
    encode_kv = Dropout(0.5)(encode_kv)
    preds = Dense(1, activation='sigmoid')(encode_kv)

    classifier = Model(inputs=input_emb, outputs=preds,
                       name='Classifier_%s' % params['kv_encoder'])
    classifier.compile(loss='mse', optimizer='sgd')

    preds = classifier(emb_kv)

    return preds, classifier
