# load packages
import os
import sys
import json
import random as python_random
import numpy as np
import tensorflow as tf

from sklearn.metrics import accuracy_score, log_loss

import keras
from keras.layers import Input, Embedding, Dense, Bidirectional, Lambda, Multiply
from keras.layers import CuDNNLSTM, Conv1D, GlobalMaxPooling1D, GlobalAveragePooling1D
from keras.layers import Dropout, concatenate
from keras.models import Model, Sequential
from keras import backend as K

from utils import attend, classify, set_trainability, DataGenerator

argv = sys.argv
try:
    gpu_idx = eval(argv[1])
except IndexError:
    gpu_idx = 0

os.environ['CUDA_VISIBLE_DEVICES'] = '%d' % gpu_idx
os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'

np.random.seed(0)
python_random.seed(0)
tf.set_random_seed(0)


def build_session(gpu_fraction=0.2, allow_growth=True):
    gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=gpu_fraction,
                                allow_growth=allow_growth)
    return tf.Session(config=tf.ConfigProto(gpu_options=gpu_options))


sess = build_session()
K.set_session(sess)
init_op = tf.group(tf.global_variables_initializer(),
                   tf.local_variables_initializer())
sess.run(init_op)

# load data
data = np.load('./data.npz')
X_kv = data['X_kv']
X_q = data['X_q']
y = data['y']
emb_matrix = data['emb']

kv_len_origin = data['kv_len_origin']
kv_len = X_kv.shape[1]
q_len = X_q.shape[1]
voab_size = emb_matrix.shape[0]
emb_dim = emb_matrix.shape[1]


# define the network

def mask_classifiy(mask):
    input_mask = Input(shape=(kv_len,), dtype='float32')

    hidden = Dense(128, activation='relu')(input_mask)
    hidden = Dense(32, activation='relu')(hidden)
    preds = Dense(1, activation='sigmoid')(hidden)

    mask_classifier = Model(
        inputs=input_mask, outputs=preds, name='MaskClassifier')
    mask_classifier.compile(loss='mse', optimizer='sgd')

    preds = mask_classifier(mask)

    return preds, mask_classifier


def get_model(params, emb_matrix):
    kv_len = params['kv_len']
    q_len = params['q_len']
    voab_size = params['voab_size']
    emb_dim = params['emb_dim']

    # encode data
    input_kv = Input(shape=(kv_len,), dtype='int32', name='Input_kv')
    input_q = Input(shape=(q_len,), dtype='int32', name='Input_q')

    emb = Embedding(voab_size, emb_dim, weights=[
                    emb_matrix], trainable=False, name='WordEmbedding')
    emb_kv = emb(input_kv)
    emb_q = emb(input_q)

    # attention
    assert params['use_attention'] == True, 'Parameter Error: use_attention'
    if params['use_attention']:
        emb_kv, mask, module_attention = attend(emb_q, emb_kv, params)

    preds, module_classifier = classify(emb_kv, params)
    mask_preds, module_mask_classifier = mask_classifiy(mask)

    def weighted_binary_crossentropy(y_true, y_pred, mask_preds=mask_preds):
        y_pred = K.constant(y_pred) if not K.is_tensor(y_pred) else y_pred
        y_true = K.cast(y_true, y_pred.dtype)

        weights = tf.stop_gradient(0.5 / ((y_true * mask_preds + (1 - y_true)
                                           * (1 - mask_preds)) * 0.99999 + 0.00001))

        return K.mean(
            K.binary_crossentropy(y_true, y_pred, from_logits=False) * weights, axis=-1)

    set_trainability(module_attention, True)
    set_trainability(module_classifier, False)
    set_trainability(module_mask_classifier, False)
    model_attention = Model(inputs=[input_kv, input_q], outputs=preds)
    model_attention.compile(loss='binary_crossentropy', optimizer='rmsprop')

    set_trainability(module_attention, False)
    set_trainability(module_classifier, True)
    set_trainability(module_mask_classifier, False)
    model_classifier = Model(inputs=[input_kv, input_q], outputs=preds)
    model_classifier.compile(
        loss=weighted_binary_crossentropy, optimizer='rmsprop')

    set_trainability(module_attention, False)
    set_trainability(module_classifier, False)
    set_trainability(module_mask_classifier, True)
    model_mask_classifier = Model(
        inputs=[input_kv, input_q], outputs=mask_preds)
    model_mask_classifier.compile(
        loss='binary_crossentropy', optimizer='rmsprop')

    if params['use_attention']:
        model_mask = Model(inputs=[input_kv, input_q], outputs=mask)
        model_mask.compile(loss='mse', optimizer='sgd')
    else:
        model_mask = None

    return model_attention, model_classifier, model_mask_classifier, model_mask, module_attention, module_classifier, module_mask_classifier


# train the model
use_attention = True
for kv_encoder in ['pooling', 'rcnn']:
    params = {
        'use_attention': use_attention,
        'kv_encoder': kv_encoder,

        'kv_len': kv_len,
        'q_len': q_len,
        'voab_size': voab_size,
        'emb_dim': emb_dim,
    }
    print(params)

    model_attention, model_classifier, model_mask_classifier, model_mask, module_attention, module_classifier, module_mask_classifier = get_model(
        params, emb_matrix)

    def train(data_gen, epochs=10):
        y_loss = []
        weight_loss = []

        for epoch in range(epochs):
            y_loss_batch = []
            weight_loss_batch = []

            set_trainability(module_attention, False)
            set_trainability(module_classifier, False)
            set_trainability(module_mask_classifier, True)
            model_mask_classifier.fit_generator(data_gen, verbose=0)

            n_step = len(data_gen)
            for step in range(n_step):
                X_batch, y_batch = data_gen[step]

                set_trainability(module_attention, False)
                set_trainability(module_classifier, False)
                set_trainability(module_mask_classifier, True)
                weight_loss_batch.append(
                    model_mask_classifier.train_on_batch(X_batch, y_batch))

                set_trainability(module_attention, True)
                set_trainability(module_classifier, False)
                set_trainability(module_mask_classifier, False)
                model_attention.train_on_batch(X_batch, y_batch)

                set_trainability(module_attention, False)
                set_trainability(module_classifier, True)
                set_trainability(module_mask_classifier, False)
                y_loss_batch.append(
                    model_classifier.train_on_batch(X_batch, y_batch))

                set_trainability(module_attention, False)
                set_trainability(module_classifier, False)
                set_trainability(module_mask_classifier, True)
                weight_loss_batch.append(
                    model_mask_classifier.train_on_batch(X_batch, y_batch))

                print('Epoch %2d/%2d. Step %3d%%. MAL: %.5f (y), %.5f (weight)' %
                      (epoch, epochs, (step + 1) / n_step * 100, np.mean(y_loss_batch), np.mean(
                          weight_loss_batch)),
                      end='\r')

                data_gen.on_epoch_end()

            y_loss.append(np.mean(y_loss_batch))
            weight_loss.append(np.mean(weight_loss_batch))

            print('')

        return y_loss, weight_loss

    data_gen = DataGenerator(X_kv, X_q, y, batch_size=128, shuffle=True)
    train(data_gen, epochs=25)

    # evaluate the model
    y_pred = model_classifier.predict([X_kv, X_q], batch_size=512).ravel()
    loss = log_loss(y, np.clip(y_pred, 1e-6, 1 - 1e-6))
    acc = accuracy_score(y, y_pred > 0.5)
    print(loss, acc)
    masks = model_mask.predict([X_kv, X_q], batch_size=128)
    neg_masks = masks[y == 0]
    pos_masks = masks[y == 1]

    result_str = {
        'kv_encoder': str(kv_encoder),
        'kv_len': int(kv_len),
        'kv_len_origin': int(kv_len_origin),
        'train_loss': float(loss),
        'train_acc': float(acc),

        'pos_default': float(np.sum(pos_masks[:, kv_len_origin:]) / np.sum(pos_masks)),
        'pos_A': float(np.sum(pos_masks[:, kv_len_origin]) / np.sum(pos_masks)),
        'pos_B': float(np.sum(pos_masks[:, kv_len_origin + 1]) / np.sum(pos_masks)),
        'pos_C': float(np.sum(pos_masks[:, kv_len_origin + 2]) / np.sum(pos_masks)),
        'pos_D': float(np.sum(pos_masks[:, kv_len_origin + 3]) / np.sum(pos_masks)),

        'neg_default': float(np.sum(neg_masks[:, kv_len_origin:]) / np.sum(neg_masks)),
        'neg_A': float(np.sum(neg_masks[:, kv_len_origin]) / np.sum(neg_masks)),
        'neg_B': float(np.sum(neg_masks[:, kv_len_origin + 1]) / np.sum(neg_masks)),
        'neg_C': float(np.sum(neg_masks[:, kv_len_origin + 2]) / np.sum(neg_masks)),
        'neg_D': float(np.sum(neg_masks[:, kv_len_origin + 3]) / np.sum(neg_masks)),
    }
    print(result_str)

    f = open('./weight.jsonl', 'a')
    f.write(json.dumps(result_str) + '\n')
    f.close()
