###########################################
# Import Packages
###########################################

import os
import time
import pickle
import numpy as np
import tensorflow as tf

from keras.datasets import mnist
from keras.preprocessing import sequence
from keras.utils import to_categorical
from keras.layers import Conv2D, Input, MaxPooling2D, Multiply, Flatten, Reshape
from keras.layers import Layer, Dense, Dropout, Activation, concatenate
from keras.models import Model, Sequential
from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau
from keras import backend as K

from utils import create_dataset_from_score, calculate_acc, Maskout

os.environ["CUDA_VISIBLE_DEVICES"] = '2'

# random seed setting
tf.set_random_seed(10086)
np.random.seed(10086)

# model setting
max_features = 5000
maxlen = 400
embedding_dims = 50
filters = 250
kernel_size = 3
hidden_dims = 250

# training setting
original_epochs = 10
pretrain_epochs = 25
expain_epochs = 10
batch_size = 64

# explanation setting
tau = 0.5
k = 25  # Number of selected pixels

###########################################
# Load Data
###########################################


def load_data():
    """
    Load data if data have been created.
    Create data otherwise.

    """

    if 'data' not in os.listdir('.'):
        os.mkdir('data')
    if 'models' not in os.listdir('.'):
        os.mkdir('models')

    if 'mnist_data.npz' not in os.listdir('data'):
        print('Loading data...')

        data_path = './data/mnist_aligned.npz'
        npz_data = np.load(data_path)
        x_train, y_train, x_val, y_val = npz_data['x_train'], npz_data['y_train'], npz_data['x_test'], npz_data[
            'y_test']

        num1 = 8
        num2 = 3
        idx = np.where(((y_train == num1) + (y_train == num2)) > 0)[0]
        x_train = x_train[idx]
        y_train = y_train[idx]
        idx = np.where(((y_val == num1) + (y_val == num2)) > 0)[0]
        x_val = x_val[idx]
        y_val = y_val[idx]

        y_train = y_train > y_train.mean()
        y_val = y_val > y_val.mean()

        x_train = x_train.astype('float32')
        x_val = x_val.astype('float32')

        if x_train.max() > 2:
            x_train /= 255
            x_val /= 255

        mx = np.mean(x_train, axis=0, keepdims=True)
        x_train -= mx
        x_val -= mx

        num_classes = len(np.unique(y_train))
        y_train = to_categorical(y_train, num_classes)
        y_val = to_categorical(y_val, num_classes)

        np.savez('./data/mnist_data.npz', x_train=x_train,
                 y_train=y_train, x_val=x_val, y_val=y_val)

    else:
        data = np.load('./data/mnist_data.npz')
        x_train = data['x_train']
        y_train = data['y_train']
        x_val = data['x_val']
        y_val = data['y_val']

    return x_train, y_train, x_val, y_val

###########################################
# Original Model
###########################################


def create_original_model(input_shape=(28, 28, 1), num_classes=2):

    model = Sequential()
    model.add(Conv2D(32, kernel_size=(3, 3),
                     activation='relu',
                     input_shape=input_shape))
    model.add(Conv2D(64, (3, 3), activation='relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dropout(0.25))
    model.add(Dense(128, activation='relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_classes, activation='softmax'))

    model.compile(loss='categorical_crossentropy',
                  optimizer='adam',
                  metrics=['accuracy'])

    return model


def generate_original_preds(train=True):

    x_train, y_train, x_val, y_val = load_data()
    input_shape = x_train[0].shape
    num_classes = y_train.shape[1]

    model = create_original_model(input_shape, num_classes)

    if train:
        filepath = "models/original.hdf5"
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True, mode='max')
        callbacks_list = [checkpoint]
        model.fit(x_train, y_train, validation_data=(x_val, y_val),
                  callbacks=callbacks_list, epochs=original_epochs, batch_size=batch_size)

    model.load_weights('./models/original.hdf5',
                       by_name=True)

    pred_train = model.predict(x_train, verbose=1, batch_size=1000)
    pred_val = model.predict(x_val, verbose=1, batch_size=1000)
    if not train:
        print('The val accuracy is {}'.format(calculate_acc(pred_val, y_val)))
        print('The train accuracy is {}'.format(
            calculate_acc(pred_train, y_train)))

    np.save('data/pred_train.npy', pred_train)
    np.save('data/pred_val.npy', pred_val)


###########################################
# Functions and Layers
###########################################

def set_trainability(model, trainable=True):
    model.trainable = trainable
    for layer in model.layers:
        layer.trainable = trainable


class Concatenate(Layer):
    """
    Layer for concatenation.

    """

    def __init__(self, **kwargs):
        super(Concatenate, self).__init__(**kwargs)

    def call(self, inputs):
        input1, input2 = inputs
        input1 = tf.expand_dims(input1, axis=-2)  # [batchsize, 1, input1_dim]
        dim1 = int(input2.get_shape()[1])
        input1 = tf.tile(input1, [1, dim1, 1])
        return tf.concat([input1, input2], axis=-1)

    def compute_output_shape(self, input_shapes):
        input_shape1, input_shape2 = input_shapes
        input_shape = list(input_shape2)
        input_shape[-1] = int(input_shape[-1]) + int(input_shape1[-1])
        input_shape[-2] = int(input_shape[-2])
        return tuple(input_shape)


class Sample_Concrete(Layer):
    """
    Layer for sample Concrete / Gumbel-Softmax variables.

    """

    def __init__(self, tau0, k, **kwargs):
        self.tau0 = tau0
        self.k = k
        super(Sample_Concrete, self).__init__(**kwargs)

    def call(self, logits):

        batch_size = K.shape(logits)[0]
        logits_ = K.permute_dimensions(
            logits, (0, 3, 1, 2))  # [batch_size, 1, h, w]

        h, w = int(logits_.get_shape()[2]), int(logits_.get_shape()[3])
        unif_shape = [batch_size, self.k, h, w]

        uniform = K.random_uniform(shape=unif_shape, minval=np.finfo(
            tf.float32.as_numpy_dtype).tiny, maxval=1.0)

        gumbel = - K.log(-K.log(uniform))
        noisy_logits = (gumbel + logits_) / self.tau0
        noisy_logits = tf.reshape(noisy_logits, [-1, self.k, h * w])
        samples = K.softmax(noisy_logits, axis=2)
        noisy_logits = tf.reshape(noisy_logits, [-1, self.k, h, w])
        samples = K.max(noisy_logits, axis=1)

        logits = tf.reshape(logits, [-1, h, w])

        output = K.in_train_phase(samples, logits)
        return tf.expand_dims(output, -1)

    def compute_output_shape(self, input_shape):
        return input_shape


def construct_gumbel_selector(X_ph, h, w, y_ph=None):

    x = X_ph
    if y_ph is not None:
        hy = Dense(h * w, activation='linear')(y_ph)
        hy = Reshape((h, w, 1))(hy)
        x = concatenate([x, hy], axis=-1)

    x = Conv2D(32, kernel_size=(3, 3),
               activation='relu', padding='same')(x)
    x = Conv2D(64, (3, 3), activation='relu', padding='same')(x)
    logits_T = Conv2D(1, (3, 3), activation='linear', padding='same')(x)

    return logits_T


def L2X(train=True):

    print('Loading dataset...')
    x_train, y_train, x_val, y_val = load_data()
    input_shape = x_train[0].shape
    num_classes = y_train.shape[1]

    pred_train = np.load('data/pred_train.npy')
    pred_val = np.load('data/pred_val.npy')
    print('Creating model...')

    # P(S|X)
    with tf.variable_scope('selection_model'):
        X_ph = Input(shape=input_shape, dtype='float32')

        logits_T = construct_gumbel_selector(
            X_ph, input_shape[0], input_shape[1])
        tau = 0.5
        T = Sample_Concrete(tau, k)(logits_T)

    # q(X_S)
    with tf.variable_scope('prediction_model'):
        x = Multiply()([X_ph, T])
        x = Conv2D(32, kernel_size=(3, 3), activation='relu')(x)
        x = Conv2D(64, (3, 3), activation='relu')(x)
        x = MaxPooling2D(pool_size=(2, 2))(x)
        x = Flatten()(x)
        x = Dense(128, activation='relu')(x)
        preds = Dense(num_classes, activation='softmax')(x)

    model = Model(inputs=X_ph,
                  outputs=preds)

    model.compile(loss='categorical_crossentropy',
                  optimizer='rmsprop',  # optimizer,
                  metrics=['acc'])
    train_acc = np.mean(np.argmax(pred_train, axis=1)
                        == np.argmax(y_train, axis=1))
    val_acc = np.mean(np.argmax(pred_val, axis=1) == np.argmax(y_val, axis=1))
    print('The train and validation accuracy of the original model is {} and {}'.format(
        train_acc, val_acc))

    if train:
        filepath = "models/l2x.hdf5"
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True, mode='max')
        callbacks_list = [checkpoint]
        st = time.time()
        model.summary()
        model.fit(x_train, pred_train,
                  validation_data=(x_val, pred_val),
                  callbacks=callbacks_list,
                  epochs=expain_epochs, batch_size=batch_size)
        duration = time.time() - st
        print('Training time is {}'.format(duration))

    model.load_weights('models/l2x.hdf5', by_name=True)

    pred_model = Model(X_ph, logits_T)
    pred_model.compile(loss='categorical_crossentropy',
                       optimizer='adam', metrics=['acc'])

    st = time.time()
    scores = pred_model.predict(x_val, verbose=1, batch_size=batch_size)
    return scores, x_val


def L2X_y(train=True):

    print('Loading dataset...')
    x_train, y_train, x_val, y_val = load_data()
    input_shape = x_train[0].shape
    num_classes = y_train.shape[1]

    pred_train = np.load('data/pred_train.npy')
    pred_val = np.load('data/pred_val.npy')
    print('Creating model...')

    # P(S|X)
    with tf.variable_scope('selection_model'):
        X_ph = Input(shape=input_shape, dtype='float32')
        y_ph = Input(shape=(2,), dtype='float32')

        logits_T = construct_gumbel_selector(
            X_ph, input_shape[0], input_shape[1], y_ph)
        tau = 0.5
        T = Sample_Concrete(tau, k)(logits_T)

    # q(X_S)
    with tf.variable_scope('prediction_model'):
        x = Multiply()([X_ph, T])
        x = Conv2D(32, kernel_size=(3, 3), activation='relu')(x)
        x = Conv2D(64, (3, 3), activation='relu')(x)
        x = MaxPooling2D(pool_size=(2, 2))(x)
        x = Flatten()(x)
        x = Dense(128, activation='relu')(x)
        preds = Dense(num_classes, activation='softmax')(x)

    model = Model(inputs=[X_ph, y_ph],
                  outputs=preds)

    model.compile(loss='categorical_crossentropy',
                  optimizer='rmsprop',  # optimizer,
                  metrics=['acc'])
    train_acc = np.mean(np.argmax(pred_train, axis=1)
                        == np.argmax(y_train, axis=1))
    val_acc = np.mean(np.argmax(pred_val, axis=1) == np.argmax(y_val, axis=1))
    print('The train and validation accuracy of the original model is {} and {}'.format(
        train_acc, val_acc))

    if train:
        filepath = "models/l2x_y.hdf5"
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True, mode='max')
        callbacks_list = [checkpoint]
        st = time.time()
        model.summary()
        model.fit([x_train, pred_train], pred_train,
                  validation_data=([x_val, pred_val], pred_val),
                  callbacks=callbacks_list,
                  epochs=expain_epochs, batch_size=batch_size)
        duration = time.time() - st
        print('Training time is {}'.format(duration))

    model.load_weights('models/l2x_y.hdf5', by_name=True)

    pred_model = Model([X_ph, y_ph], logits_T)
    pred_model.compile(loss='categorical_crossentropy',
                       optimizer='adam', metrics=['acc'])

    st = time.time()
    scores = pred_model.predict(
        [x_val, pred_val], verbose=1, batch_size=batch_size)
    return scores, x_val


def L2X_pretrain_y(train=True):

    print('Loading dataset...')
    x_train, y_train, x_val, y_val = load_data()
    input_shape = x_train[0].shape
    num_classes = y_train.shape[1]

    pred_train = np.load('data/pred_train.npy')
    pred_val = np.load('data/pred_val.npy')
    print('Creating model...')

    def make_explainer(infer_input, y_input):
        logits_T = construct_gumbel_selector(
            infer_input, input_shape[0], input_shape[1], y_input)
        tau = 0.5
        T = Sample_Concrete(tau, k)(logits_T)

        return T, logits_T

    def make_approximator(infer_input,
                          T,
                          pretrain_input):

        net = Sequential([
            Conv2D(32, kernel_size=(3, 3), activation='relu'),
            Conv2D(64, (3, 3), activation='relu'),
            MaxPooling2D(pool_size=(2, 2)),
            Flatten(),
            Dense(128, activation='relu'),
            Dense(num_classes, activation='softmax')
        ])

        infer_output = net(Multiply()([infer_input, T]))
        pretrain_output = net(
            Maskout(keep_rate=float(k) / 784)(pretrain_input))

        return infer_output, pretrain_output

    # make all
    infer_input = Input(shape=input_shape, dtype='float32')
    y_input = Input(shape=(2,), dtype='float32')
    T, logits_T = make_explainer(infer_input, y_input)

    pretrain_input = Input(shape=input_shape, dtype='float32')
    infer_output, pretrain_output = make_approximator(
        infer_input, T, pretrain_input)

    # compile the model
    explainer = Model([infer_input, y_input], logits_T)
    explainer.compile(loss='categorical_crossentropy',
                      optimizer='adam', metrics=['acc'])

    pretrain_apporximator = Model(pretrain_input, pretrain_output)
    pretrain_apporximator.compile(loss='categorical_crossentropy',
                                  optimizer='rmsprop',  # optimizer,
                                  metrics=['acc'])

    set_trainability(pretrain_apporximator, False)
    e2e_model = Model([infer_input, y_input], infer_output)
    e2e_model.compile(loss='categorical_crossentropy',
                      optimizer='rmsprop',  # optimizer,
                      metrics=['acc'])
    set_trainability(pretrain_apporximator, True)

    train_acc = np.mean(np.argmax(pred_train, axis=1)
                        == np.argmax(y_train, axis=1))
    val_acc = np.mean(np.argmax(pred_val, axis=1) == np.argmax(y_val, axis=1))
    print('The train and validation accuracy of the original model is {} and {}'.format(
        train_acc, val_acc))

    if train:
        filepath = "models/L2X_pretrain_y.hdf5"
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True, mode='max')
        reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.2,
                                      patience=2, min_lr=0.00001, verbose=1)
        st = time.time()
        pretrain_apporximator.summary()
        pretrain_apporximator.fit(x_train, pred_train,
                                  validation_data=(x_val, pred_val),
                                  callbacks=[checkpoint, reduce_lr],
                                  epochs=pretrain_epochs, batch_size=batch_size)
        pretrain_apporximator.load_weights(
            'models/L2X_pretrain_y.hdf5', by_name=True)
        set_trainability(pretrain_apporximator, False)

        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True, mode='max')
        e2e_model.summary()
        e2e_model.fit([x_train, pred_train], pred_train,
                      validation_data=([x_val, pred_val], pred_val),
                      callbacks=[checkpoint],
                      epochs=expain_epochs, batch_size=batch_size)
        duration = time.time() - st
        print('Training time is {}'.format(duration))

    e2e_model.load_weights('models/L2X_pretrain_y.hdf5', by_name=True)

    st = time.time()
    scores = explainer.predict(
        [x_val, pred_val], verbose=1, batch_size=batch_size)
    return scores, x_val


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--task', type=str,
                        choices=['original', 'L2X', 'L2X_y', 'L2X_pretrain_y'], default='original')
    parser.add_argument('--train', action='store_true')
    parser.set_defaults(train=False)
    args = parser.parse_args()
    dict_a = vars(args)

    if args.task == 'original':
        generate_original_preds(args.train)
    elif args.task == 'L2X':
        scores, x = L2X(args.train)
        create_dataset_from_score(x, scores, k, name='L2X')
    elif args.task == 'L2X_y':
        scores, x = L2X_y(args.train)
        create_dataset_from_score(x, scores, k, name='L2X_y')
    elif args.task == 'L2X_pretrain_y':
        scores, x = L2X_pretrain_y(args.train)
        create_dataset_from_score(x, scores, k, name='L2X_pretrain_y')
