###########################################
# Import Packages
###########################################

import os
import sys
import time
import pickle
import numpy as np
import pandas as pd
import tensorflow as tf

from keras.preprocessing import sequence
from keras.utils import to_categorical
from keras.preprocessing.sequence import pad_sequences
from keras.preprocessing.text import Tokenizer
from keras.layers import Conv1D, Input, GlobalMaxPooling1D, Multiply, Lambda, Embedding
from keras.layers import Layer, Dense, Dropout, Activation, concatenate
from keras.models import Model, Sequential
from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau
from keras import backend as K

from utils import create_dataset_from_score, calculate_acc, Maskout

os.environ["CUDA_VISIBLE_DEVICES"] = '1'

# random seed setting
tf.set_random_seed(10086)
np.random.seed(10086)

# model setting
max_features = 5000
maxlen = 200
embedding_dims = 50
filters = 250
kernel_size = 3
hidden_dims = 250

# training setting
original_epochs = 5
pretrain_epochs = 20
expain_epochs = 5
batch_size = 400

# explanation setting
tau = 0.5
k = 5  # Number of selected words


###########################################
# Load Data
###########################################

def load_data():

    if 'data' not in os.listdir('.'):
        os.mkdir('data')
    if 'models' not in os.listdir('.'):
        os.mkdir('models')

    if 'id_to_word.pkl' not in os.listdir('./data'):
        train_data = pd.read_csv('./data/train.csv', header=None)
        test_data = pd.read_csv('./data/test.csv', header=None)

        data = pd.concat([train_data, test_data])
        data.columns = ['label', 'text']

        X = data['text'].values
        X = ['startofsentence ' + sent for sent in X]
        y = to_categorical(data['label'].values - 1)

        tokenizer = Tokenizer(num_words=5000, oov_token='<oov>')
        tokenizer.fit_on_texts(X)
        word_to_id = tokenizer.word_index
        id_to_word = {value: key for key, value in word_to_id.items()}
        id_to_word[word_to_id['startofsentence']] = '<start>'

        X = tokenizer.texts_to_sequences(X)
        X = pad_sequences(X, maxlen=200)

        x_train, x_val = X[:len(train_data)], X[len(train_data):]
        y_train, y_val = y[:len(train_data)], y[len(train_data):]

        np.savez('./data/yelp_data.npz', x_train=x_train,
                 y_train=y_train, x_val=x_val, y_val=y_val)
        with open('./data/id_to_word.pkl', 'wb') as f:
            pickle.dump(id_to_word, f)
    else:
        data = np.load('./data/yelp_data.npz')
        x_train = data['x_train']
        y_train = data['y_train']
        x_val = data['x_val']
        y_val = data['y_val']
        with open('./data/id_to_word.pkl', 'rb') as f:
            id_to_word = pickle.load(f)

    return x_train, y_train, x_val, y_val, id_to_word


###########################################
# Original Model
###########################################

def create_original_model():

    model = Sequential()
    model.add(Embedding(max_features,
                        embedding_dims,
                        input_length=maxlen))
    model.add(Dropout(0.2))
    model.add(Conv1D(filters,
                     kernel_size,
                     padding='valid',
                     activation='relu',
                     strides=1))
    model.add(GlobalMaxPooling1D())
    model.add(Dense(hidden_dims))
    model.add(Dropout(0.2))
    model.add(Activation('relu'))
    model.add(Dense(2))
    model.add(Activation('softmax'))

    model.compile(loss='categorical_crossentropy',
                  optimizer='adam',
                  metrics=['accuracy'])

    return model


def generate_original_preds(train=True):

    x_train, y_train, x_val, y_val, id_to_word = load_data()
    model = create_original_model()

    if train:
        filepath = './models/original.hdf5'
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True)
        callbacks_list = [checkpoint]
        model.fit(x_train, y_train, validation_data=(x_val, y_val),
                  callbacks=callbacks_list, epochs=original_epochs, batch_size=batch_size)

    model.load_weights('./models/original.hdf5')

    pred_train = model.predict(x_train, verbose=1, batch_size=1000)
    pred_val = model.predict(x_val, verbose=1, batch_size=1000)
    if not train:
        print('The val accuracy is {}'.format(calculate_acc(pred_val, y_val)))
        print('The train accuracy is {}'.format(
            calculate_acc(pred_train, y_train)))

    np.save('data/pred_train.npy', pred_train)
    np.save('data/pred_val.npy', pred_val)


###########################################
# Functions and Layers
###########################################

Mean = Lambda(lambda x: K.sum(x, axis=1) / float(k),
              output_shape=lambda x: [x[0], x[2]])


def set_trainability(model, trainable=True):

    model.trainable = trainable
    for layer in model.layers:
        layer.trainable = trainable


class Concatenate(Layer):

    def __init__(self, **kwargs):
        super(Concatenate, self).__init__(**kwargs)

    def call(self, inputs):
        input1, input2 = inputs
        input1 = tf.expand_dims(input1, axis=-2)  # [batchsize, 1, input1_dim]
        dim1 = int(input2.get_shape()[1])
        input1 = tf.tile(input1, [1, dim1, 1])
        return tf.concat([input1, input2], axis=-1)

    def compute_output_shape(self, input_shapes):
        input_shape1, input_shape2 = input_shapes
        input_shape = list(input_shape2)
        input_shape[-1] = int(input_shape[-1]) + int(input_shape1[-1])
        input_shape[-2] = int(input_shape[-2])
        return tuple(input_shape)


class Sample_Concrete(Layer):

    def __init__(self, tau0, k, **kwargs):
        self.tau0 = tau0
        self.k = k
        super(Sample_Concrete, self).__init__(**kwargs)

    def call(self, logits):
        # logits: [batch_size, d, 1]
        logits_ = K.permute_dimensions(logits, (0, 2, 1))  # [batch_size, 1, d]

        d = int(logits_.get_shape()[2])
        unif_shape = [batch_size, self.k, d]

        uniform = K.random_uniform_variable(shape=unif_shape,
                                            low=np.finfo(
                                                tf.float32.as_numpy_dtype).tiny,
                                            high=1.0)
        gumbel = - K.log(-K.log(uniform))
        noisy_logits = (gumbel + logits_) / self.tau0
        samples = K.softmax(noisy_logits)
        samples = K.max(samples, axis=1)
        logits = tf.reshape(logits, [-1, d])
        threshold = tf.expand_dims(tf.nn.top_k(
            logits, self.k, sorted=True)[0][:, -1], -1)
        discrete_logits = tf.cast(
            tf.greater_equal(logits, threshold), tf.float32)

        output = K.in_train_phase(samples, discrete_logits)
        return tf.expand_dims(output, -1)

    def compute_output_shape(self, input_shape):
        return input_shape


def construct_gumbel_selector(X_ph, y_ph, num_words, embedding_dims, maxlen):

    first_layer = Embedding(num_words, embedding_dims,
                            input_length=maxlen)(X_ph)  # (400, 50)
    first_layer = Conv1D(100, kernel_size, padding='same',
                         activation='relu', strides=1)(first_layer)

    # global info
    global_info = GlobalMaxPooling1D(
        name='new_global_max_pooling1d_1')(first_layer)
    if y_ph is not None:
        y_info = Dense(64, activation='relu')(y_ph)
        global_info = concatenate([global_info, y_info])
    global_info = Dense(100, name='new_dense_1',
                        activation='relu')(global_info)

    # local info
    local_info = Conv1D(100, 3, padding='same',
                        activation='relu', strides=1)(first_layer)
    local_info = Conv1D(100, 3, padding='same',
                        activation='relu', strides=1)(local_info)

    # combined info
    combined_info = Concatenate()([global_info, local_info])
    combined_info = Dropout(0.2, name='new_dropout_2')(combined_info)
    combined_info = Conv1D(100, 1, padding='same',
                           activation='relu', strides=1)(combined_info)

    logits_T = Conv1D(1, 1, padding='same', activation=None,
                      strides=1)(combined_info)
    T = Sample_Concrete(tau, k)(logits_T)

    return logits_T, T


def construct_approximator(infer_ph, pretrain_ph=None):

    network = Sequential([
        Dropout(0.2),
        Conv1D(filters, kernel_size, padding='valid',
               activation='relu', strides=1),
        GlobalMaxPooling1D(),
        Dense(hidden_dims),
        Dropout(0.2),
        Activation('relu'),
        Dense(2),
        Activation('softmax'),
    ])

    if pretrain_ph is None:
        return network(infer_ph)
    else:
        return network(infer_ph), network(pretrain_ph)


###########################################
# Explaination Methods
###########################################

def L2X(train=True, use_y=False):

    print('Loading dataset...')
    x_train, y_train, x_val, y_val, id_to_word = load_data()
    pred_train = np.load('data/pred_train.npy')
    pred_val = np.load('data/pred_val.npy')

    print('Creating model...')

    # input placeholder
    X_ph = Input(shape=(maxlen,), dtype='int32')
    if not use_y:
        y_ph = None
    else:
        y_ph = Input(shape=(2,), dtype='float32')

    # explainer
    logits_T, T = construct_gumbel_selector(
        X_ph, y_ph, max_features, embedding_dims, maxlen)

    # approximator
    appr = Embedding(max_features, embedding_dims, input_length=maxlen)(X_ph)
    appr = Multiply()([appr, T])
    preds = construct_approximator(appr)

    # make model
    if not use_y:
        model = Model(inputs=X_ph, outputs=preds)
        explainer = Model(X_ph, logits_T)
        filepath = './models/l2x.hdf5'
    else:
        model = Model(inputs=[X_ph, y_ph], outputs=preds)
        explainer = Model([X_ph, y_ph], logits_T)
        filepath = './models/l2x_y.hdf5'

        x_train = [x_train, pred_train]
        x_val = [x_val, pred_val]

    model.compile(loss='categorical_crossentropy',
                  optimizer='rmsprop',
                  metrics=['acc'])
    explainer.compile(loss='categorical_crossentropy',
                      optimizer='rmsprop', metrics=['acc'])

    # train the model
    if train:
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True)

        st = time.time()
        model.summary()
        model.fit(x_train, pred_train,
                  validation_data=(x_val, pred_val),
                  callbacks=[checkpoint],
                  epochs=expain_epochs, batch_size=batch_size)

        duration = time.time() - st
        print('Training time is {}'.format(duration))

    model.load_weights(filepath)

    # select features
    st = time.time()
    scores = explainer.predict(
        x_val, verbose=1, batch_size=batch_size)[:, :, 0]

    if not use_y:
        return scores, x_val
    else:
        return scores, x_val[0]


def L2X_pretrain(train=True, use_y=True):

    print('Loading dataset...')
    x_train, y_train, x_val, y_val, id_to_word = load_data()
    pred_train = np.load('data/pred_train.npy')
    pred_val = np.load('data/pred_val.npy')

    print('Creating model...')

    # input placeholder
    infer_ph = Input(shape=(maxlen,), dtype='int32')
    pretrain_ph = Input(shape=(maxlen,), dtype='int32')
    if not use_y:
        y_ph = None
    else:
        y_ph = Input(shape=(2,), dtype='float32')

    # explainer
    logits_T, T = construct_gumbel_selector(
        infer_ph, y_ph, max_features, embedding_dims, maxlen)

    # approximator
    emb = Embedding(max_features, embedding_dims, input_length=maxlen)
    infer_input = Multiply()([emb(infer_ph), T])
    pretrain_input = Maskout(keep_rate=k / maxlen,
                             noise_shape=(maxlen, 1))(emb(pretrain_ph))
    infer_preds, pretrain_preds = construct_approximator(
        infer_input, pretrain_input)

    # make models
    if not use_y:
        explainer = Model(infer_ph, logits_T)
        pretrain_apporximator = Model(pretrain_ph, pretrain_preds)
        e2e_model = Model(infer_ph, infer_preds)
        filepath = './models/l2x_pretrain.hdf5'
    else:
        explainer = Model([infer_ph, y_ph], logits_T)
        pretrain_apporximator = Model([pretrain_ph, y_ph], pretrain_preds)
        e2e_model = Model([infer_ph, y_ph], infer_preds)
        filepath = './models/l2x_pretrain_y.hdf5'

        x_train = [x_train, pred_train]
        x_val = [x_val, pred_val]

    explainer.compile(loss='categorical_crossentropy',
                      optimizer='rmsprop', metrics=['acc'])
    pretrain_apporximator.compile(loss='categorical_crossentropy',
                                  optimizer='rmsprop', metrics=['acc'])
    set_trainability(pretrain_apporximator, False)
    e2e_model.compile(loss='categorical_crossentropy',
                      optimizer='rmsprop',
                      metrics=['acc'])
    set_trainability(pretrain_apporximator, True)

    if train:
        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True)
        reduce_lr = ReduceLROnPlateau(monitor='val_acc', factor=0.2,
                                      patience=2, min_lr=0.00001, verbose=1)
        st = time.time()
        pretrain_apporximator.summary()
        pretrain_apporximator.fit(x_train, pred_train,
                                  validation_data=(x_val, pred_val),
                                  callbacks=[checkpoint, reduce_lr],
                                  epochs=pretrain_epochs, batch_size=batch_size)
        pretrain_apporximator.load_weights(filepath)
        set_trainability(pretrain_apporximator, False)

        checkpoint = ModelCheckpoint(filepath, monitor='val_acc',
                                     verbose=1, save_best_only=True)
        e2e_model.summary()
        e2e_model.fit(x_train, pred_train,
                      validation_data=(x_val, pred_val),
                      callbacks=[checkpoint],
                      epochs=expain_epochs, batch_size=batch_size)
        duration = time.time() - st
        print('Training time is {}'.format(duration))

    e2e_model.load_weights(filepath)

    st = time.time()
    scores = explainer.predict(
        x_val, verbose=1, batch_size=batch_size)[:, :, 0]

    if not use_y:
        return scores, x_val
    else:
        return scores, x_val[0]


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--task', type=str,
                        choices=['original', 'L2X', 'L2X_y', 'L2X_pretrain_y'], default='original')
    parser.add_argument('--train', action='store_true')
    parser.set_defaults(train=False)
    args = parser.parse_args()
    dict_a = vars(args)

    if args.task == 'original':
        generate_original_preds(args.train)
    elif args.task == 'L2X':
        scores, x = L2X(args.train)
        create_dataset_from_score(x, scores, k, name='L2X')
    elif args.task == 'L2X_y':
        scores, x = L2X(args.train, use_y=True)
        create_dataset_from_score(x, scores, k, name='L2X_y')
    elif args.task == 'L2X_pretrain_y':
        scores, x = L2X_pretrain(args.train, use_y=True)
        create_dataset_from_score(x, scores, k, name='L2X_pretrain_y')
